/*
* $Log$
* Revision 2.1  2004/02/18 19:26:17  richq
* Armies either belong to a character or to a fortress/city (analogous to
* garrisons in LoM) Loyalty, race, size and energy depend on the owner of the army
* and whether it is in a Fortress or "travelling" with the character.
*
*/

#include "stdlib.h"
#include "gba.h"
#include "army.h"
#include "main.h"
#include "characters.h"
#include "tokendata.h"
#include "tokens.h"

int g_garrisonHere;
garrison_t * g_garrisons;
chararmy_t * g_charArmies;

#define TABLE_SIZE    128
#define ARMY_G_SIZE             (TABLE_SIZE * 7)
#define ARMY_G_OWNER            (TABLE_SIZE * 9)
#define ARMY_C_SIZE             (TABLE_SIZE * 6)
#define ARMY_C_LOST             (TABLE_SIZE * 5)
#define ARMY_C_SLEW             (TABLE_SIZE * 4)
#define ARMY_C_RACE             (TABLE_SIZE * 8)

void init_army(void) {
  if (g_charArmies == 0) {
    g_garrisons  = (garrison_t *)malloc(128 * sizeof(garrison_t));
    g_charArmies = (chararmy_t *)malloc(128 * sizeof(chararmy_t));
    
//    gbaprintf("sizeof(garrison_t) %d",sizeof(garrison_t));
//    gbaprintf("sizeof(chararmy_t) %d",sizeof(chararmy_t));
  }
  
  int i;
  for (i = 0; i < 128; i++) {
    g_garrisons[i].size  = tablesInRom[i+ARMY_G_SIZE];        // table 7
    g_garrisons[i].owner = tablesInRom[i+ARMY_G_OWNER];       // table 9
    
    g_charArmies[i].size = tablesInRom[i+ARMY_C_SIZE];        // table 6
    g_charArmies[i].lost = tablesInRom[i+ARMY_C_LOST];        // table 5
    g_charArmies[i].slew = tablesInRom[i+ARMY_C_SLEW];        // table 4
    g_charArmies[i].race = tablesInRom[i+ARMY_C_RACE];        // table 8
  }
  
}

int getArmyLoyalty(int armyid) {
  
  // character armies' loyalty will be the same as the character
  int owner = armyid;
  if (armyid > 128) {
    // is fortress army, 
    // their loyalty is the same as the character who owns the army
    owner = g_garrisons[armyid-128].owner;
  }
  return LOYALTY(g_characters[owner].flags);
  
}
int getArmyRace(int armyid) {
  
  // character armies' race taken straight from army
  int owner = armyid;
  if (armyid > 128) {
    // is fortress garrison, their race is the same 
    // as the army of the character who owns the garrison
    owner = g_garrisons[armyid-128].owner;
  }
  return g_charArmies[owner].race;
  
}

int getArmySize(int armyid) {
  if (armyid > 128)
    return g_garrisons[armyid-128].size;
  else
    return g_charArmies[armyid].size;
}

void setArmySize(int armyid, int newSize) {
  if (armyid > 128)
    g_garrisons[armyid-128].size = newSize;
  else
    g_charArmies[armyid].size = newSize;
}


int getArmyEnergy(int armyid) {
  if (armyid > 128)
    return 180; // garrisons always have this energy 
  else
    return g_characters[armyid].energy; // character armies are as tired as character
}



void calculateGarrisonAt(const location_t * pos) {
  // 79d8
  int i; // fortress id
  int loyalTo;
  for (i = 0; i < 128; i++) {
    if (g_fortresses[i].x == pos->x && g_fortresses[i].y == pos->y) {
      // we are stood at a fortress
      g_garrisonHere = i+128;
      loyalTo = getArmyLoyalty(g_garrisonHere);
      if (g_status.loyalty != loyalTo) {
        g_unloyals_here++;
      }
      return;
    }
    
  }
  g_garrisonHere = 0;
}


void addArmyTypeTokens(int army) {
  int race = getArmyRace(army);
  addSingleToken((race&1) + 0x70);
}


//; army race name table
//; 76 = moonguard
//; 77 = ice guard
//; 23 = fey
//; 24 = barbarian
// 773f 
const u8 armyRaceNames[] = {
  0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x23, 0x23, 
  0x24, 0x24, 0x77, 0x77, 0x00, 0x25, 0x77, 0x26
};

void addArmySizeAndType(int id) {
  numberToWords(getArmySize(id)*5); // "one hundred"
  if (getArmySize(id)) {
    addUppercaseToken();
    
    addSingleToken(armyRaceNames[getArmyRace(id)]);
    addArmyTypeTokens(id);
    
  } else {
    addSingleToken(0x7a); // "army"
  }
  addLiteralToken('.');
}
 


void createCharArmyReport(int who) {
  // 770c
  createCharacterTitle(who);
  addSingleToken(0x6f); //"commands"
  addArmySizeAndType(who); // no army/one hundred barbarian warriors
  
  if (getArmySize(who)==0)
    return;
  
  // get army status
  addToTokenList(theyAre);
  addUtterly(getArmyEnergy(who));
  addToTokenList(tired_);
  
  
}

