/*
* $Log$
* Revision 2.3  2004/02/22 15:17:04  richq
* Added code to get just friendly chars at a location. Also guidance
* is now recorded.
*
* Revision 2.2  2004/02/18 19:26:54  richq
* Character info is not held in tables as per the speccy version -
* instead easier-to-follow structs are used.
* Routines here for finding who is at a location, character tiredness,
* checking the current status, recruiting, and creating the "character
* reports" shown on info screens.
*
* Revision 2.1.1.1  2004/02/04 21:49:21  Default
* Doomdark's Revenge GBA
*
*
*/


#ifndef _characters_h

#define _characters_h

#include "gba.h"
#include "location.h"


#define LUXOR 0
#define MORKIN 1
#define TARITHEL 2
#define SHARETH 4

#define DAWN 0x1f

#define RACE_LUXOR_HORSE     0
#define RACE_LUXOR           1
#define RACE_MORKIN_HORSE    2
#define RACE_MORKIN          3
#define RACE_RORTHRON_HORSE  4
#define RACE_RORTHRON        5
#define RACE_FEY_HORSE       6
#define RACE_FEY             7
#define RACE_BARBARIAN_HORSE 8
#define RACE_BARBARIAN       9
#define RACE_ICELORD_HORSE   10
#define RACE_ICELORD         11
#define RACE_TARITHEL        12
#define RACE_GIANT           13
#define RACE_SHARETH         14
#define RACE_DWARF           15

#define KILLED_FOE_FLAG   8
#define IN_BATTLE_FLAG    16
#define WON_BATTLE_FLAG   32
#define IN_TUNNEL_FLAG    64
#define USED_OBJECT_FLAG  128

// for character.flags variable:
#define LOYALTY(flags)     (flags&7  )
#define KILLED_FOE(flags)  (flags&KILLED_FOE_FLAG  )
#define IN_BATTLE(flags)   (flags&IN_BATTLE_FLAG )
#define WON_BATTLE(flags)  (flags&WON_BATTLE_FLAG )
#define IN_TUNNEL(flags)   ((flags&IN_TUNNEL_FLAG ) != 0)
#define USED_OBJECT(flags) (flags&USED_OBJECT_FLAG)
// for looking/time
#define LOOKING(var) (var&7)
#define MASK_LOOKING(var) (var&0xf8)
#define TIME_LEFT(var)  (MASK_LOOKING(var)>>3)
// for orders
#define GET_ORDERS(race_orders) ((race_orders>>4)&0xf)
typedef struct {
  location_t location;          // 0xa100, 0xa180
  u8 time_looking; // flag 1    // 0xa200
  u8 race_orders;  // flag 2    // 0xa280
  u8 flags;        // flag 3    // 0xa300
  u8 energy;                    // 0xa380
  u8 fighting;                  // 0xa400
  u8 foe;                       // 0xa480
  u8 liege;                     // 0xa500
  u8 reckless;                  // 0xa580
  u8 object;                    // 0xa600
  u8 attribute_good;            // 0xa680
  u8 attribute_bad;             // 0xa700
  u8 despondence;                   // 0xa780
}
character_t;

// good attributes:
#define IS_GOOD(val)         ((val &  1    )!=0)
#define IS_STRONG(val)       ((val & (1<<1))!=0)
#define IS_FORCEFUL(val)     ((val & (1<<2))!=0)
#define IS_GENEROUS(val)     ((val & (1<<3))!=0)
#define IS_STUBBORN(val)     ((val & (1<<4))!=0)
#define IS_BRAVE(val)        ((val & (1<<5))!=0)
#define IS_SWIFT(val)        ((val & (1<<6))!=0)
#define IS_LOYAL(val)        ((val & (1<<7))!=0)
// bad attributes
#define IS_EVIL(val)         ((val &  1    )!=0)
#define IS_WEAK(val)         ((val & (1<<1))!=0)
#define IS_RETICENT(val)     ((val & (1<<2))!=0)
#define IS_GREEDY(val)       ((val & (1<<3))!=0)
#define IS_FAWNING(val)      ((val & (1<<4))!=0)
#define IS_COWARDLY(val)     ((val & (1<<5))!=0)
#define IS_SLOW(val)         ((val & (1<<6))!=0)
#define IS_TREACHEROUS(val)  ((val & (1<<7))!=0)

extern int g_unloyals_here;
extern int g_total_here;

extern u8 * g_charactersHere;

#define tablesInRom _binary_tables_bin_start
extern u8 tablesInRom[]; // can I do this? tablesInRom[][128]


//extern int g_selectedCharID;
//extern character_t * g_selectedChar;
extern character_t * g_characters;


extern int g_gameover;


//void setSelectedChar(int i);
void init_characters(void);
void getCharactersAt(const location_t * pos);

void getAllFriendsAt(const location_t * pos);

void increaseU8By(u8 * val, int increment);
void decreaseU8By(u8 * val, int decrement);
void makeCharacterTired(int tire);

int checkLocationAndStatus(void);
int doApproach(int them, int us);

// involves a lot of characters stuff
void describeGuidance(void);

void describeObjectCarried(int who);
void addPreparesToBattle(int who);
void checkDeadChar(int who);
void describeCharacter(int who);

void addCharAndKiller(int who);

#endif
