/*
* $Log$
* Revision 2.3  2004/02/22 15:24:45  richq
* Changed const ints to their actual sizes. Layer register fiddling to
* try and get smoother screen transitions.
*
* Revision 2.2  2004/02/18 19:25:54  richq
* Mostly unchanged from LoM. Differences are the addition of mist,
* the way armies are shown and the specific table values for
* feature positions.
*
* Revision 2.1.1.1  2004/02/04 21:49:20  Default
* Doomdark's Revenge GBA
*
*
*/
#include "screenmode.h"
#include "gba.h"
#include "gfx.h"
#include "location.h"
#include "characters.h"
#include "features.h"
#include "location.h"
#include "savearea.h"
#include "blit.h"

// first the look up tables:
typedef struct {
u8 dy;    // location y adjuster
s8 dx;    // location x adjuster
s16 xadj;  // screen x adjuster
u8 y;     // screen y position
u8 size;  // scale
} landscape_t;

// Landscaping Location Calculation Table
// lives at 9843 in DR
const landscape_t landscape[] = {
     {5,4,86,72,5   },
     {4,5,114,72,5  },
     {6,2,41,72,5   },
     {6,1,21,72,5   },
     {6,-1,-21,72,5 },
     {6,0,0,72,5    },
     {5,3,69,71,5   },
     {3,5,131,71,5  },
     {4,4,100,71,5  },
     {5,2,48,71,5   },
     {5,1,25,71,5   },
     {5,-1,-25,71,5 },
     {5,0,0,70,5    },
     {4,3,82,70,5   },
     {3,4,118,70,5  },
     {4,2,59,70,5   },
     {3,3,100,69,4  },
     {4,1,31,69,4   },
     {4,-1,-31,69,4 },
     {4,0,0,68,4    },
     {3,2,75,67,4   },
     {2,3,125,67,4  },
     {3,1,41,66,3   },
     {3,-1,-41,66,3 },
     {3,0,0,65,3    },
     {2,2,100,65,3  },
     {2,1,59,61,2   },
     {1,2,141,61,2  },
     {2,0,0,59,2    },
     {1,1,100,51,1  },
     {1,0,0,40,0    }
};



// table from 0xc000. Indices into 0xc0cc
const u16 terrain[][8] = {
  { 0x0000, 0x0211, 0x0383, 0x048a, 0x053a, 0x05ba, }, // Mountains
  { 0x0616, 0x0814, 0x0971, 0x0a5a, 0x0aea, 0x0b52, }, // Forest   
  { 0x0ba0, 0x0c65, 0x0cf0, 0x0d4f, 0x0d89, 0x0db5, }, // Hills    
  { 0x0dd8, 0x0ec7, 0x0f69, 0x0fd7, 0x101f, 0x104d, }, // Gate     
  { 0x1070, 0x1200, 0x1319, 0x13da, 0x145e, 0x14c1, }, // Temple   
  { 0x1510, 0x1596, 0x15f3, 0x162b, 0x164c, 0x1664, }, // Pit      
  { 0x1676, 0x17e3, 0x18d1, 0x1965, 0x19b5, 0x19f2, }, // Palace   
  { 0x1a2a, 0x1b38, 0x1bef, 0x1c66, 0x1cb4, 0x1ce6, }, // Fortress 
  { 0x1d0c, 0x1dd6, 0x1e62, 0x1ec2, 0x1f02, 0x1f30, }, // Hall     
  { 0x1f55, 0x1fdf, 0x2034, 0x2065, 0x2080, 0x2095, }, // Hut      
  { 0x20a4, 0x2222, 0x2323, 0x23d8, 0x2442, 0x2493, }, // Tower    
  { 0x24c8, 0x28f7, 0x2b9c, 0x2d12, 0x2dde, 0x2e4d, }, // City     
  { 0x2e90, 0x2fa3, 0x304a, 0x30a8, 0x30e1, 0x3104, }, // Fountain 
  { 0x3121, 0x3268, 0x334f, 0x33ee, 0x344a, 0x3489, }, // Stones   
  { 0x34ae, 0x362d, 0x3732, 0x37e4, 0x3848, 0x388e, }, // Wastes   
  { 0x38c5, 0x39ca, 0x3a82, 0x3b07, 0x3b58, 0x3b8e  }, // Army
  { 0x3bb5, 0x3cf6, 0x3dcd, 0x3e5e, 0x3ea5, 0x3ed2  }  // Mist
};                                                        


int getLocation ( int x, int y ) {
  if (x < 0 || x > 0x3f || y < 0 || y >= 0x60)
    return WASTES;
  return g_map[x+(y)*64];
}

int AdjustMapPos ( int pos, int type, int dx, int dy )
{
   switch ( type ) {
       case 0: pos = pos + dy; break;
       case 1: pos = pos - dy; break;
       case 2: pos = pos + dx; break;
       case 3: pos = pos - dx; break;
   }
   return pos ;
}

void drawLocation(int distance, int xtype, int ytype, int scrtype ) {
  int locx,locy;
  int   x;
  // work out the real y coordinate
  // i.e. y coordinate of this location
  locy = AdjustMapPos (  g_characters[g_saveArea.selectedChar].location.y,
                         ytype,
                         landscape[distance].dx,
                         landscape[distance].dy );

  // work out the real x coordinate
  locx = AdjustMapPos (  g_characters[g_saveArea.selectedChar].location.x,
                         xtype,
                         landscape[distance].dx,
                         landscape[distance].dy );

  // actually get the location data...
  // use map for this
  
  int feature = getLocation ( locx, locy ) ;

  // work out where on screen, x coordinate this item should be
  x = 320;
  switch ( scrtype ) {
    // case 0 = right side, looking straight
    case 0: x += landscape[distance].xadj ;          break;
    // case 1 = left side, looking straight
    case 1: x -= landscape[distance].xadj ;          break;
    // case 2 = right side, looking diagonal
    case 2: x = (x+landscape[distance].xadj)-100 ;   break;
    // case 3 = left side, looking diagonal
    case 3: x = (x-landscape[distance].xadj)+100 ;   break;
  }

  // and draw it
  drawFeature ( feature&0xf,
                landscape[distance].size,
                x,
                landscape[distance].y);
  
  if ( HAS_MIST(feature) ) {
    drawFeature( 16,
                landscape[distance].size,
                x - (0x4c-landscape[distance].y),
                landscape[distance].y);
    drawFeature( 16,
                landscape[distance].size,
                x + (0x4c-landscape[distance].y),
                landscape[distance].y);  
  }
  else if ( HAS_ARMY(feature) ) {
    drawFeature( 17,
    landscape[distance].size,
    x - (0x4c-landscape[distance].y),
    landscape[distance].y);
    
    drawFeature( 17,
    landscape[distance].size,
    x + (0x4c-landscape[distance].y),
    landscape[distance].y);  
  }
  
  
}


// Landscaping Direction Adjustment tables
// used in AdjustMapPos:
// 0: pos = pos + dy; 
// 1: pos = pos - dy; 
// 2: pos = pos + dx; 
// 3: pos = pos - dx; 
// dx, dy come from landscape table at the given distance
const u8 xAdjustments[] = { 3,2,0,0,2,3,1,1,3 };
const u8 yAdjustments[] = { 1,1,3,2,0,0,2,3,1 };
void drawPanorama(int dir) {
  
  // clear out the tiles..
  
  clearBuffer();
  
  // looking along a straight line
  int howLftScr = 1;
  int howRhtScr = 0;
  // adjust for diagonal
  if (dir&1 ) {
    howLftScr = 2;
    howRhtScr = 3;
  }
  int ii;
  // step through all the entries in the landscaping table
  for (ii=0; ii<31; ii++ ) {
    // draw location on the left of the screen
    drawLocation( ii,
                  xAdjustments[dir],
                  yAdjustments[dir],
                  howLftScr );

    // draw locations on the right of the screen  
    drawLocation( ii,
                  xAdjustments[dir+1],
                  yAdjustments[dir+1],
                  howRhtScr );
  }
  //wait_vsync_int();
  if (g_panning == PAN_RIGHT) {
    // pan right... scroll left
    scrollBufferLeft();
  }
  else if (g_panning == PAN_LEFT) {
    // pan left, scroll right
    scrollBufferRight();
  } else {
    REG_DISPCNT &= ~BG0_ENABLE;
//    clear_view();
    blit_buffer();
    wait_vsync_int();
    setup_view();
    
    
  }
  g_panning = 0;
}

void drawFeature(int feature, int size, int xi, int yi) {
    
  if (feature == 0)
    return;
  
  feature--;
  
  // max height of feature is 78 pixels
  // min value of y offset is 40
  // 78+40 = 118
  int y = (118 - yi);
  
  int x = xi - 256;
  int index = terrain[feature][size];
  
  int height = featureData[index++];
  int y_offset = 0;
  
  int startx, endx;
  while ( height-- != 0) {
    y_offset = y*256;
    for (;;) {
      
      int flags = featureData[index++];
      // bit 7 = More than 1 group of drawing operations
      // for this scan line
      // bit 6 = Should we draw in ink or paper
      // bits 0 - 5 = No of drawing instructions to follow
      
      int moreOnLine = (flags & 0x80);
      int inkOrPaper = (flags & 0x40) ? 0 : 1 ;   // 0 = paper, 1 = ink
      int opCount = flags & 0x3f;

      // draw a horizontal line from x1 to x2
      int x1 = featureData[index++];
      int x2 = featureData[index++];
      startx = x1;
      endx = x2;
      drawLineToBuffer(x+x1, x+x2, y_offset, 
                       inkOrPaper?BACKGROUND_INK:BACKGROUND_PAPER);
      
      // draw on the edges in Ink Always
      drawSetToBuffer(x+x1, y_offset, BACKGROUND_INK);
      drawSetToBuffer(x+x2, y_offset, BACKGROUND_INK);

      // reverse the last color
      inkOrPaper = (inkOrPaper+1)&1;
      while ( (opCount--) != 0 ) {
        // bit 7 = 1 = Draw A Line from
        // bits 0 - 6 = x coordinate
        flags = featureData[index++];
        x1 = flags & 0x7f ;
        if (flags&0x80) {
          x2 = featureData[index++];
          drawLineToBuffer(x+x1, x+x2, y_offset, 
                           inkOrPaper?BACKGROUND_INK:BACKGROUND_PAPER);

        }
        else {
          drawSetToBuffer(x+x1, y_offset, 
                          inkOrPaper?BACKGROUND_INK:BACKGROUND_PAPER);
        }
      
      }
      if ( !moreOnLine ) break;
    }
    y--;
  }
  
  
}
