/*
* $Log$
* Revision 2.4  2006/07/30 11:43:31  rich
* Align with current gbalib (remove common sprite routines)
* fix malloc warning
*
* Revision 2.3  2004-02-22 15:28:04  richq
* Added the top windows' lower edge.
* Initialisation of backgrounds loses its copy-pasted nature.
*
* Revision 2.2  2004/02/18 19:19:17  richq
* Mostly the same as LoM. New things are:
*  extra palette swapping and tunnel stuff.
*  scrolling view has to take into account that the mist and army gfx
*   are "hacked in" - they overlap the calculated view when very close.
*
* Revision 2.1.1.1  2004/02/04 21:49:20  Default
* Doomdark's Revenge GBA
*
*
*/

#include <malloc.h>
#include "dma.h"
#include "gba.h"
#include "sprite.h"
#include "screenmode.h"

#include "inter.h"

#include "gfx.h"
#include "stdlib.h"

#include "memory_defs.h"

int g_panningOn = 1;
int g_panning;

#define HORIZON_SCROLL (256-41-16)

#define FONT_COL_INDEX 15

// tile data for the sky, ground and fonts starts in char base block 1
// char base block 0 holds the main "features" graphics
#define HORIZON_SKY 1
#define HORIZON_GROUND 2
#define TOP_SCREEN_TILE 3
#define TOP_SCREEN_BASE_TILE 4
#define TOP_SCREEN_BASE_TILE_2 5
#define HORIZON_SKY_4  (HORIZON_SKY|(HORIZON_SKY<<4)|(HORIZON_SKY<<8)|(HORIZON_SKY<<12))
#define HORIZON_GROUND_4  (HORIZON_GROUND|(HORIZON_GROUND<<4)|(HORIZON_GROUND<<8)|(HORIZON_GROUND<<12))
#define TOP_SCREEN_4 (FONT_COL_INDEX|(FONT_COL_INDEX<<4)|(FONT_COL_INDEX<<8)|(FONT_COL_INDEX<<12))
#define TOP_SCREEN_BASE_4 HORIZON_GROUND_4


#define HORIZON_SKY_PAL (16+1)
#define HORIZON_GROUND_PAL (16+2)

int g_horizonSky;
int g_horizonGround;

u8 * g_buffer;// = (u8*)ADDR_G_BUFFER;

Bg bg_level;
Bg bg_text;
Bg bg_horizon;

OAMEntry sprites[128];
unsigned char fadedup;

void calculate_skyline(void);

// gamma correction = sqrt(col*4)
// bright is 223  -> 30  
// non bright is 191 ->
#define BRIGHT 30
#define NONBRT 27
const u16 brightPal[] = {
  RGB16(0,0,0),                //black,
  RGB16(0,0,BRIGHT),           //blue,
  RGB16(BRIGHT,0,0),           //red,
  RGB16(BRIGHT,0,BRIGHT),      // purple
  RGB16(0,BRIGHT,0),           //green,
  RGB16(0,BRIGHT,BRIGHT),      // cyan 
  RGB16(BRIGHT,BRIGHT,0),      // yellow
  RGB16(BRIGHT,BRIGHT,BRIGHT), // white
};
const u16 nonBrightPal[] = {
  RGB16(0,0,0),                 //black,
  RGB16(0,0,NONBRT),            //blue,
  RGB16(NONBRT,0,0),            //red,
  RGB16(NONBRT,0,NONBRT),       // purple
  RGB16(0,NONBRT,0),            //green,
  RGB16(0,NONBRT,NONBRT),       // cyan 
  RGB16(NONBRT,NONBRT,0),       // yellow
  RGB16(NONBRT,NONBRT,NONBRT),  // white
};

// clears out the video memory, which may have been set already by the boot menu

void clear_view(void) {
  
  bg_level.x_scroll = 8;
  int x,y;
  // sets the view field
  for (x = 0; x < 32; x++) {
    for (y = 0; y < 10; y++) {
      bg_level.mapData[x+y*32] = 0;
    }
  }
  
}

void clear_text_from(int startY) {
  int x,y;
  for (x = 0; x < 32; x++) {
    for (y = startY; y < 20; y++) {
      bg_text.mapData[x+y*32] = 0;
    }
  }
}

void clear_text(void) {
  clear_text_from(0);
}



void setup_view(void) {
  int x, y;
  
  int i = 1;
  for (y=0; y < 10; y++) {
    for (x=0; x < 32; x++) {
      bg_level.mapData[x+y*32] = i;
      i++;
    }
  }
  
  bg_level.x_scroll = 8;
  bg_level.y_scroll = HORIZON_SCROLL;
  
  UpdateBackground(&bg_level);
  REG_DISPCNT |= BG0_ENABLE;
}


void init_backgrounds(void) {

  // create a bg  
  initBackground(&bg_level, 0, 0, 28, 1);

  
  initBackground(&bg_text, 1, 1, 29, 0);

  int x, y;
  for (y=0; y < 20; y++) {
    for (x=0; x < 30; x++) {
      bg_text.mapData[x+y*32] = 0;
    }
  }
  
  initBackground(&bg_horizon, 2, 1, 30, 2);
  
  
}

void set_think_palette(void) {
  set_horizon();
  BGPaletteMem[BACKGROUND_INK]   = RGB16(0,0,0);
  BGPaletteMem[BACKGROUND_PAPER] = RGB16(NONBRT,NONBRT,NONBRT);
  BGPaletteMem[HORIZON_SKY_PAL] = RGB16(NONBRT,NONBRT,NONBRT);
  BGPaletteMem[HORIZON_GROUND_PAL] = RGB16(NONBRT,NONBRT,NONBRT);
  int x,y;
  for (y=0; y < 7; y++) {
    for (x=0; x < 32; x++) {
        bg_horizon.mapData[x+y*32] = HORIZON_GROUND|(1<<12);
    }
  }
}


void set_night_palette(void) {
  BGPaletteMem[HORIZON_SKY_PAL]   = RGB16(0,0,0);
  BGPaletteMem[HORIZON_GROUND_PAL] = RGB16(0,0,0);
  int x,y;
  for (y=0; y < 7; y++) {
    for (x=0; x < 32; x++) {
        bg_horizon.mapData[x+y*32] = HORIZON_GROUND|(1<<12);
    }
  }
}

void set_flame_tile(int x, int y) {
  bg_horizon.mapData[x+y*32] = HORIZON_GROUND|(1<<12);
}

int l_horizon_set = 0;
void set_tunnel_palette(void) {
  // a hacky way to get the flames right
  g_horizonSky = 0;
  g_horizonGround = brightPal[6];
  
  calculate_skyline();
  int x,y;
  for (y=7; y < 20; y++) {
    for (x=0; x < 32; x++) {
        bg_horizon.mapData[x+y*32] = 0; 
    }
  }
  l_horizon_set = 0;
}


void set_horizon_palette(void) {
  BGPaletteMem[BACKGROUND_INK] = RGB16(0,0,BRIGHT);
  BGPaletteMem[BACKGROUND_PAPER] = RGB16(0,BRIGHT,BRIGHT);

  g_horizonSky = brightPal[1];
  g_horizonGround = brightPal[5];

}

void set_palette_time(int time) {
  
  // this would be a good place to turn on HBLANK interrupts...
  if (time == 0) {
    // night
    wait_vsync_int();
    BGPaletteMem[BACKGROUND_INK]   = 0x0;
    BGPaletteMem[BACKGROUND_PAPER] = brightPal[1];
    
    g_horizonSky = 0;
    g_horizonGround = brightPal[1];

  }else
  if (time == 0x1f) {
    // dawn..
    wait_vsync_int();
    BGPaletteMem[BACKGROUND_INK] =   nonBrightPal[6];
    BGPaletteMem[BACKGROUND_PAPER] = nonBrightPal[2];
      
    g_horizonSky = nonBrightPal[6];
    g_horizonGround = nonBrightPal[2];

  }
  else {
    set_horizon_palette();  
  }
  calculate_skyline();
}
 

void set_horizon(void) {
  if (l_horizon_set) return;
  
  // else must be leaving a tunnel...
  BGPaletteMem[HORIZON_SKY_PAL]   = 0;
  BGPaletteMem[HORIZON_GROUND_PAL] = 0;
  
  //set_horizon_palette();
  
  int x, y;
  
  for (y=0; y < 7; y++) {
    for (x=0; x < 32; x++) {
      bg_horizon.mapData[x+y*32] = TOP_SCREEN_TILE|(FONT_WHITE_B<<12);
      
    }
  }
  for (y=7; y < 13; y++) {
    for (x=0; x < 32; x++) {
      bg_horizon.mapData[x+y*32] = HORIZON_SKY|(1<<12);
    }
  }
  for (y= 0; y < 8; y++) {
    for (x=0; x < 32; x++) {
      bg_horizon.mapData[13*32+x+y*32] = HORIZON_GROUND|(1<<12);
    }
  }
  int i = 0;
  int offset = HORIZON_SKY*16;
  // this does the stripes.
  for (i = 0; i < 16; ) {
    bg_horizon.tileData[offset+i] = HORIZON_SKY_4;
    i++;
    bg_horizon.tileData[offset+i] = HORIZON_SKY_4;
    i++;
    bg_horizon.tileData[offset+i] = HORIZON_GROUND_4;
    i++;
    bg_horizon.tileData[offset+i] = HORIZON_GROUND_4;
    i++;
    
  }
  
  
  offset = HORIZON_GROUND*16;
  for (i = 0; i < 32; i++)
    bg_horizon.tileData[offset+i] = HORIZON_GROUND_4;
  
  offset = TOP_SCREEN_TILE*16;
  for (i = 0; i < 16; i++)
    bg_horizon.tileData[offset+i] = TOP_SCREEN_4;
  
  offset = TOP_SCREEN_BASE_TILE*16;
  for (i = 0; i < 16; i++)
    bg_horizon.tileData[offset+i] = TOP_SCREEN_4;
  for (i = 12; i < 16; i++)
    bg_horizon.tileData[offset+i] = TOP_SCREEN_BASE_4;
  offset = TOP_SCREEN_BASE_TILE_2*16;
  for (i = 0; i < 16; i++)
    bg_horizon.tileData[offset+i] = TOP_SCREEN_4;
  for (i = 8; i < 10; i++)
    bg_horizon.tileData[offset+i] = TOP_SCREEN_BASE_4;
  for (i = 12; i < 16; i++)
    bg_horizon.tileData[offset+i] = TOP_SCREEN_BASE_4;
  
  setup_view();
  l_horizon_set = 1;
}


const unsigned short paperMasks[] = {
  (FONT_COL_INDEX),     //0x000F,   
  (FONT_COL_INDEX<<4),  //0x00F0,   
  (FONT_COL_INDEX<<8),  //0x0F00,   
  (FONT_COL_INDEX<<12)  //0xF000    
};

void fade_up(void) {
  u8 loop;
  
  REG_BLDMOD = BLDMOD_FIRST_BG1 | BLDMOD_FIRST_BG0 | BLDMOD_FIRST_BG2 |
               BLDMOD_FIRST_BG3 |
               BLDMOD_FIRST_OBJ | BLDMOD_SECOND_BD | BLDMOD_MODE(3);
  
  for (loop = 0; loop < 17; loop+=2) {
    wait_vsync_int();
    REG_COLEY = (16-loop);
  }
  wait_vsync_int();
  REG_COLEY = 0;
  fadedup = 1;
}     

void fade_down(void) {
  u8 loop;
  
  REG_BLDMOD = BLDMOD_FIRST_BG1 | BLDMOD_FIRST_BG0 | BLDMOD_FIRST_BG2 |
               BLDMOD_FIRST_BG3 |
               BLDMOD_FIRST_OBJ | BLDMOD_SECOND_BD | BLDMOD_MODE(3);
  
  for (loop = 0; loop < 17; loop+=2) {
    wait_vsync_int();
    REG_COLEY = loop;
  }
  wait_vsync_int();
  REG_COLEY = 16;
  fadedup = 0;
}


// 39478 in sna, da1b in memory
#define fontGFX _binary_font_bin_start 
extern u8 fontGFX[];
// load the font data into tile ram
void load_font(void) {
  // the font is stored in 1 bit per pixel format
  // expand it to 4bpp
  // each letter is 8*8, 4bpp = 8*4 pixels = 32 u8s per char
  
  int i, j, k, m;
  
  // amount to shift by
  int shift; 
  // four pixels, 4 bits each
  u16 pixelQuad; 
  // start index in tile ram
  u32 address = FONT_TILE*16;
  // offset in fontGfX array of the character
  int offset = 0;
  u8 theByte;
  u8 bitMask;
  for (j = 0; j < 95; j++) {
    //offset = j*8;
    // loop over characters
    for (i = 0; i < 8; i++) {
      // loop over rows
      theByte = fontGFX[offset++];
      if (theByte != 0) {
        // has data
        // 66
        for (k = 0; k < 2; k++) {
          // each row requires 2 writes to vram
          
          pixelQuad = (u16)0x0000;
          // create the u16 of 4 pixels:
          for (m = 0; m < 4; m++) {
            bitMask = 1;
            // pixel k*4+m on this row
            shift = k*4 + m;
            
            if (shift != 7)
              bitMask <<= (7-shift);
            
            if ( (theByte & bitMask) != 0) {
//              if (m == 0)
//                pixelQuad = FONT_COL_INDEX;
//              else
              pixelQuad |= paperMasks[m];//(FONT_COL_INDEX<<(m*4));
            }
          }
          bg_text.tileData[address++] = pixelQuad;
          
        }
      } else {
        // the byte is empty, draw a blank row
        for (m = 0; m < 2; m++) {
          bg_text.tileData[address++] = 0x0000; // pixel quad of 00 00
        }
      }
    }
  }  
}

// print character at row, col in text bg
// use palette pal
void print_text(char text, int row, int col, int pal) {
  if (text != 0x0) {
    bg_text.mapData[row+col*32] = (FONT_TILE + (text)-0x20)|((pal&0xf)<<12);
  }
  
}



void move_sprite(int id, int x, int y) {
  if (id < 128) {
    MoveSprite(&sprites[id], x, y);
    CopyOAM();
  }
}



/*All sprites are 64*64, col256 so can "hard code" OAM entry calculations*/
u32 get_oam64(int id) {
  return id * 64*64/2;
}

void init_sprite64(int id, int priority) {
  sprites[id].attribute0 = SQUARE|COLOR_256;
  sprites[id].attribute1 = SIZE_64;
  sprites[id].attribute2 = PRIORITY(priority) | (get_oam64(id)/16);
  // OAM = 16 * sprite entry
}

const u8 * gfxTable;
void createTile(u8 * pixels, int tile, int pen, int paper) {
  
  const u8 * data = &gfxTable[tile*8];
  // read the image from offset...
  // the image is made up of 1 bit per pixel graphics
  // so for an image of 16*16, you would need 16 rows of 2 bytes
  
  int width = 1;
  int x, y, n;
  for (y = 0; y < 8; y++) {
    for (x = 0; x < width; x++) {
      // read in the byte of gfx data...
      // now expand to 8 pixels
      for (n = 0; n < 8; n++) {
        int pixelIndex = (x*8 + n) + y*8; /* *width; */
        if (n != 7)
          pixels[pixelIndex] = *data & (1<<(7-n));
        else
          pixels[pixelIndex] = *data & 1;
          
        if (pixels[pixelIndex] == 0) 
          pixels[pixelIndex] = paper;
        else
          pixels[pixelIndex] = pen;
      }
      data++;
    }
  }
  
}


// copy tile pixels into vram for this sprite at the given row/columb
// for the given sprite - all sprites are 64*64 so this is allowed
void drawImage(const u8 * pixels, u8 col, u8 row, u8 sprite) { 
  // to work out tile, use the old x+y*width formula
  // width is 8, byt the way
  int i =  col + row*8;// tile count;
  int nOAM = get_oam64(sprite);
  REG_DMA3SAD = (u32)pixels; // 8 bit reads into gfx array
  REG_DMA3DAD = (u32)&OAMData[nOAM+i*32]; // 16 bit reads into VRAM
  REG_DMA3CNT = 32 |DMA_16NOW; // 16 bit transfers from gfx array via DMA
  
}

// assume sprite is 64*64!!
void clearSpriteRAM(u8 sprite) {
  u32 zero = 0;
  int nOAM = get_oam64(sprite);
  REG_DMA3SAD = (u32)&zero; // 8 bit reads into gfx array
  REG_DMA3DAD = (u32)&OAMData[nOAM]; // 16 bit reads into VRAM
  REG_DMA3CNT = 32*4*8 |DMA_SOURCE_FIXED|DMA_32NOW; // 16 bit transfers from gfx array via DMA
  
}

//
//int rStart;
//int gStart;
//int bStart;
//           
//int rEnd;  
//int gEnd;  
//int bEnd;  
//           
//int rStep; 
//int gStep; 
//int bStep; 

void calculate_skyline(void) {
  
  BGPaletteMem[HORIZON_SKY_PAL]   = g_horizonSky;
  BGPaletteMem[HORIZON_GROUND_PAL] = g_horizonGround;
  if (g_horizonGround != brightPal[6])
    BGPaletteMem[HORIZON_GROUND+FONT_WHITE_B*16] = g_horizonGround;
  else
    BGPaletteMem[HORIZON_GROUND+FONT_WHITE_B*16] = 0;
  int x,y;
  for (y=0; y < 7; y++) {
    for (x=0; x < 32; x++) {
      bg_horizon.mapData[x+y*32] = TOP_SCREEN_TILE|(FONT_WHITE_B<<12);
    }
  }
  
  y = 6;
  for (x=0; x < 25; x++) {
    bg_horizon.mapData[x+y*32] = TOP_SCREEN_BASE_TILE|(FONT_WHITE_B<<12);
  }
  for (x=0; x < 18; x++) {
    bg_horizon.mapData[x+y*32] = TOP_SCREEN_BASE_TILE_2|(FONT_WHITE_B<<12);
  }
}


void init_gfxbuffer(void) {
  
  // init text colours too...
  if (g_buffer == 0)
    g_buffer = (u8*)malloc(300*80);
  
  int i;
  // init the bright text palettes. Skip black.
  for (i = 1; i < 8; i++) {
    BGPaletteMem[16*i+FONT_COL_INDEX] = brightPal[i];
  }
  // And the non bright ones...
  for (i = 8; i < 16; i++) {
    BGPaletteMem[16*i+FONT_COL_INDEX] = nonBrightPal[i-8];
  }
}

void clearBuffer(void) {
  u32 zero = 0;
  REG_DMA3SAD = (u32)&zero; // 8 bit reads into gfx array
  REG_DMA3DAD = (u32)g_buffer; // 16 bit reads into VRAM
  REG_DMA3CNT = (256*80/4) |DMA_SOURCE_FIXED|DMA_32NOW; // 16 bit transfers from gfx array via DMA
}


void drawLineToBuffer(int x1, int x2, int yOffset, int colour) {
  if ( x1 < 0 ) x1 = 0;
  if ( x2 > 255) x2 = 255;
  if (x1 > x2)
    return;
  int i;
  
  for (i = x1; i <= x2; i++)
    g_buffer[yOffset +i] = colour;
}

void drawSetToBuffer(int x, int yOffset, int colour) {
  if ( x<0 ) return;
  if ( x>255) return;
  g_buffer[yOffset+x] = colour;
}


// direction 
// 0 = right
// 1 = left
void scrollBuffer(int direction) {
  int i, j, scroller;
  u8 * buffer;
  u16 * pAddress;
  u16 val;
  int thisColumn;
  
  // "Carry" is set if x_scroll/8 leaves a half or not - i.e. if 
  // the buffer is currently misaligned w.r.t tile mem 
  int carry = bg_level.x_scroll&4;
  
  // When carry NOT set copy one less columns on next scroll
  // In DR mist overlaps panoramas, so 13 (12) columns are 
  // repeated, leaving 18 new ones.
  // the extra 5 columns comes from 76-40 / 8 = 4.5
  // see features, drawLocation, if ( HAS_MIST(feature) ) for the reason 
  // clue: smallest landscape[distance].y is 40 and 0x4c = 76
  // i.e. if cary set, copy an extra column 
  int max_val = (carry)?19:18;
  // When carry NOT set, need to start source at buffer+4 
  // This is because the buffer is 4 pixels out compared to current panorama
  int bufferOffset = (carry)?0:4;
  
  for (scroller = 0; scroller < max_val; scroller++) {
    // get the column to start the copying at...
    if (direction) {
      //scroll right
      if (scroller < 5) // copying to cover army/mist overlap
        thisColumn = (bg_level.x_scroll/8+25+scroller);
      else  // copying new background
        thisColumn = (bg_level.x_scroll/8+30);
    }
    else {
      if (scroller < 5) // copying to cover army/mist overlap
        thisColumn = (bg_level.x_scroll/8+(4-scroller));
      else // copying new background
        thisColumn = (bg_level.x_scroll/8+0xff);
    }
    
    if (carry && (scroller == max_val-1) ) {
      // last column copy is 1 past the scroll position
      if (direction) {
        thisColumn++;
      } else {
        thisColumn--;
      }
    }
    
    thisColumn &= 0x1f;
    
    // destination in vram for this column (+16 as the 0th tile is left blank)
    pAddress = &bg_level.tileData[16+ thisColumn*16];
    // source tile data for this column in ram
    // 104 represents the 13 columns of repeated data.
    // 136 represents the 17 columns of new data
    if (direction) {
      buffer = &g_buffer[104 + bufferOffset+scroller*8];
    } else {
      buffer = &g_buffer[136 + bufferOffset-scroller*8];
    }
    
    // 10 tiles in each column, copy them:
    for (j = 0; j < 10; j++) {
      
      // this copies 1 tile:
      for (i = 0; i < 8; i++) {
        
        // first 4 pixels on this row:
        val  = (*buffer++);
        val |= (*buffer++)<<4;
        val |= (*buffer++)<<8;
        val |= (*buffer++)<<12;
        
        *pAddress++ = val;
        
        // next 4 pixels on this row:
        val  = (*buffer++);
        val |= (*buffer++)<<4;
        val |= (*buffer++)<<8;
        val |= (*buffer++)<<12;
        
        *pAddress++ = val;
        
        // increase buffer 1 row:
        buffer+=248;
      }
      // increase tile vram pointer a row, to next tile in this column:
      pAddress += 31*16;
    }
    
    // only scroll 12 times - 12*8 = 96 pixels
    // first 5 column copies are to cover the mist overlap
    // occasional last column copy is to cover the rounding between 100 and 104
    if (scroller > 4 && scroller < 17) {
      wait_vsync_int();
      // scroll left/right by doing +8/-8
      if (direction)
        bg_level.x_scroll += 8;
      else
        bg_level.x_scroll -= 8;
      
      bg_level.x_scroll &= 0xff;
      UpdateBackground(&bg_level);
    }
  }
  // final shuffle left half a tile: Total scroll is 100 pixels
  // this 4 pixel mis-alignment is what causes all the problems
  if (direction)
    bg_level.x_scroll += 4;
  else
    bg_level.x_scroll -= 4;
    
  bg_level.x_scroll &= 0xff;
  UpdateBackground(&bg_level);
  
}

void scrollBufferLeft(void) {
  scrollBuffer(1);
}

void scrollBufferRight(void) {
  scrollBuffer(0);
}

