/*
* $Log$
* Revision 2.3  2006/07/30 11:45:42  rich
* Set initial values to 0
*
* Revision 2.2  2004/02/22 15:15:43  richq
* Finalish version of the map.
*
* Revision 2.1  2004/02/18 19:29:37  richq
* A map. Makes life a hell of a lot easier. This is not based on
* the speccy game, natch.
*
*/

#include "gba.h"
#include "screenmode.h"
#include "dma.h"
#include "keypad.h"
#include "tokens.h"
#include "tokendata.h"
#include "backgrounds.h"
#include "location.h"
#include "gfx.h"
#include "blit.h"
#include "mapscreen.h"
#include "features.h"
#include "characters.h"
#include "menu.h"
#include "main.h"
#include "savearea.h"
#include "gfx.h"
#include "objects.h"
#include "objectdata.h"

#define MAP_WIDTH 64
#define MAP_HEIGHT 96

#define MAP_FEATURE_X_POS  0x160
#define MAP_FEATURE_Y_POS  40

#define MAX_CHARS_TO_SHOW 1
#define CHAR_NAME_Y 7

#define MAP_MOVE_DELAY  4
#define MAP_A_DELAY     10
#define MAP_CURSOR_SPRITE  7

// this is bright red in the object pal:
#define MAP_CURSOR_COLOUR  (32+3)

// keep track of what we have seen:
u8 * g_haveSeenMap;

// delay in frames before info is shown (60 frames = 1 second)
#define TIMER_DELAY (60*1)

// tile indices for the info background frame
#define CORNER_TILE 0x11
#define VERTICAL_TILE 0x12
#define HORIZONTAL_TILE 0x13

#define INFO_PALETTE 0x2
#define TUNNEL_PALETTE 0x3
#define BLANK_PALETTE 0x4

// the graphics data for the mini-map tiles and the frame
#define worldMapTileData _binary_map_raw_start
extern u8 worldMapTileData [];

// used to keep track of where we have scrolled to
int g_worldX=0;
int g_worldY=0;
// the position of the cursor in world coords
location_t g_cursorLocation;
// the position of the cursor on screen
location_t g_cursorScreen={0,0};

// Map is shown on hw backgound 3
Bg bg_map;
// the transparent white panel is background 2 (replaces the horizon)
Bg bg_panel;

void copyColumn(int worldx, int worldy, int dir);
void copyRow(int worldx, int worldy, int dir);


void copyMapTilesToVideo(void) {
  int i;
  // copy tile data to vram - in 16 colour format
  // the 16 terrain types and the panel frame
  i = 0;
  REG_DMA3SAD = (u32)&worldMapTileData[i*32]; // 32 bytes * 2 pixels per byte = 64 pixels
  REG_DMA3DAD = (u32)&(bg_map.tileData[(i+1)*16]); // 16 bit reads into VRAM
  REG_DMA3CNT = ((HORIZONTAL_TILE+1)*16) |DMA_16NOW; // 16 bit transfers from gfx array via DMA 16*4 = 64
  
  // and make sure tile 0 is blank:  
  REG_DMA3SAD = (u32)&i; // 32 bytes * 2 pixels per byte = 64 pixels
  REG_DMA3DAD = (u32)&(bg_map.tileData[0]); // 16 bit reads into VRAM
  REG_DMA3CNT = (16) |DMA_16NOW|DMA_SOURCE_FIXED; // 16 bit transfers from gfx array via DMA 16*4 = 64
}

void fillMapSquare(int bgIndex, int worldx, int worldy) {
  // fill in the map square based on what we have explored
  // when a tunnel is here uses palette 3 to highlight.
  int tunnel, worldIndex;
  
  worldIndex = worldx+worldy*MAP_WIDTH;
  tunnel = HAS_TUNNEL(g_map[worldIndex]);
  
  if (HAVE_EXPLORED_AREA(worldIndex)) {
    bg_map.mapData[bgIndex] = 1+(g_map[worldIndex]&0xf);
  } else {
    // set to empty, which is really a black plains tile.
    bg_map.mapData[bgIndex] = 1 | (BLANK_PALETTE<<12);
  }
  
  if (tunnel && HAVE_EXPLORED_TUNNEL(worldIndex)) {
    // we have been in this tunnel...highlight in tunnel colour
    bg_map.mapData[bgIndex] = (bg_map.mapData[bgIndex]&0xf) | (TUNNEL_PALETTE<<12);
  } 
//  else {
    // don't know if there are tunnels here or not. leave as is.
    // bg_map.mapData[bgIndex] &= 0xFF;
//  }
  
}

void haveSeen(const location_t * pos) {
  // we have seen this square:
  int worldIndex = pos->x+pos->y*MAP_WIDTH;
  if (g_status.charInTunnel) {
    g_haveSeenMap[worldIndex] |= EXPLORED_TUNNEL;
  } else {
    g_haveSeenMap[worldIndex] |= EXPLORED_AREA;
  }
}


void haveExplored(const location_t * stood, const location_t * ahead) {
  haveSeen(stood);
  int aheadIndex = ahead->x+ahead->y*MAP_WIDTH;
  if (  g_status.canDrawInfront &&
       (g_status.charInTunnel ||
          (!g_status.charInTunnel && !HAS_MIST(g_map[aheadIndex]))
       )
       ) 
  {
    // can only see ahead if it isn't misty or if looking at a tunnel there
    haveSeen(ahead);
  }
  
}


void loadMapTiles(void) {
  u16 i, j;
  int index;
  
  // turn off the backgrounds until we need them
  REG_DISPCNT &= ~BG2_ENABLE;
  REG_DISPCNT &= ~BG3_ENABLE;
  
  // init the background object for the map, bg 3.
  // char base 0 is the feature data, char base 1 is the font data
  // so use char base 2. priority 3 means "below everything" in this case
  initBackground(&bg_map, 3, 2, 10, 3);
  // panel uses same char base, different map base and has higher priority:
  initBackground(&bg_panel, 2, 2, 11, 2);
  
  copyMapTilesToVideo();
  
  // initialise the palettes:
  // this is the cyan/blue colour for the majority of the tiles:
  BGPaletteMem[2] = brightPal[5];
  BGPaletteMem[1] = brightPal[1];
  // this is the white for the info panel
  BGPaletteMem[INFO_PALETTE*16 + 2] = brightPal[7];
  BGPaletteMem[INFO_PALETTE*16 + 1] = 0;
  // this is the green/black for the tunnels
  BGPaletteMem[TUNNEL_PALETTE*16 + 2] = RGB16(8, 25, 7);
  BGPaletteMem[TUNNEL_PALETTE*16 + 1] = 0;
  // blank is used to hide the things we don't know about...
  BGPaletteMem[BLANK_PALETTE*16 + 2] = 0;
  BGPaletteMem[BLANK_PALETTE*16 + 1] = 0;
  
  
  // now fill the initial map data...
  for (i = 0; i < 30; i++) {
    for (j = 0; j < 20; j++) {
      index = i + 32*j;
      fillMapSquare(index, i, j); 
    }
  }
  
  // this is the top leftcorner index
  g_worldX = -1;
  g_worldY = -1;
  
}

// scroll the screen and check if new tiles are used.
void scroll_down(void) {
  if ((g_worldY + 21) < MAP_HEIGHT) {
    bg_map.y_scroll+=8;
    bg_map.y_scroll &= 255;
    g_worldY++;
    copyRow(g_worldX+1,g_worldY+20,1);
    
    UpdateBackground(&bg_map);
  }
}
void scroll_up(void) {
  if (g_worldY > -1 ) {
    
    bg_map.y_scroll-=8;
    bg_map.y_scroll &= 255;
    
    g_worldY--;
    copyRow(g_worldX+1,g_worldY+1,0);
    
    UpdateBackground(&bg_map);
  }
}
void scroll_left(void) {
  if (g_worldX > -1) {
    bg_map.x_scroll-=8;
    bg_map.x_scroll &= 255;
    
    g_worldX--;
    copyColumn(g_worldX+1,g_worldY+1,0);
    
    UpdateBackground(&bg_map);
  }
}
void scroll_right(void) {
  if ((g_worldX + 30) < 63) {
    bg_map.x_scroll+=8;
    bg_map.x_scroll &= 255;
    
    g_worldX++;
    copyColumn(g_worldX+30,g_worldY+1,1);
    
    UpdateBackground(&bg_map);
    
  }
}

// copy a column into map data at the right of screen
// x = x column, y = y position
void copyColumn(int worldx, int worldy, int dir) {
  int j;
  
  // get the column just off screen:
  int theColumn;
  if (dir)
    theColumn = (bg_map.x_scroll/8+29)&31;
  else
    theColumn = (bg_map.x_scroll/8+32)&31;
  
  int yPos;
  int index;
  for (j = 0; j < 32; j++) {
    // get the y position in the bg, row 0 may not be at the top:
    yPos = ((bg_map.y_scroll/8 + j)&31);
    index = yPos*32+theColumn;
    fillMapSquare(index, worldx, worldy+j);
  }
  
}

// copy a row into map data at the right of screen
// x = x column, y = y position
void copyRow(int worldx, int worldy, int dir) {
  int j;
  
  // get the row just off screen:
  int theRow;
  if (dir)
    theRow = (bg_map.y_scroll/8+19)&31;
  else
    theRow = (bg_map.y_scroll/8+32)&31;
  
  int xPos;
  int index;
  for (j = 0; j < 32; j++) {
    // get the x position in the bg, column 0 may not be the leftmost column
    xPos = ((bg_map.x_scroll/8 + j)&31);
    index = theRow*32+xPos;
    fillMapSquare(index, worldx+j, worldy);
  }

}


// draws the white panel with a frame
// if tile = 0, then the panel is cleared.
void drawPanelBG(int panelXOffset, int tile) {
  // disable the bg...
  REG_DISPCNT &= ~BG2_ENABLE;
  
  int x, y;
  for (x = 0; x < 32; x++) { //
    for (y = 0; y < 32; y++) {
      bg_panel.mapData[x+y*32] = 0;
    }
  }
  
  if (tile == 0) {
    return;
  }
  
  // fil in the middle:
  for (x = 0; x < 15; x++) {
    for (y = 0; y < 20; y++) {
      bg_panel.mapData[x+y*32] = tile|(INFO_PALETTE<<12);
    }
  }
  // Fill in the frame. Top edge:
  for (x = 1; x < 14; x++)
    bg_panel.mapData[x] = HORIZONTAL_TILE|(INFO_PALETTE<<12);
  //Bottom edge:
  for (x = 1; x < 14; x++)
    bg_panel.mapData[x+19*32] = HORIZONTAL_TILE|TILE_FLIP_VERT|(INFO_PALETTE<<12);
  //Left edge:
  for (y = 1; y < 19; y++)
    bg_panel.mapData[0+y*32] = VERTICAL_TILE|(INFO_PALETTE<<12);
  //Right edge:
  for (y = 1; y < 19; y++)
    bg_panel.mapData[14+y*32] = VERTICAL_TILE|TILE_FLIP_HORZ|(INFO_PALETTE<<12);
  
  bg_panel.mapData[0]  = CORNER_TILE|(INFO_PALETTE<<12);
  bg_panel.mapData[14]  = CORNER_TILE|TILE_FLIP_HORZ|(INFO_PALETTE<<12);
  bg_panel.mapData[0+19*32] = CORNER_TILE|TILE_FLIP_VERT|(INFO_PALETTE<<12);
  bg_panel.mapData[14+19*32] = CORNER_TILE|TILE_FLIP_VERT|TILE_FLIP_HORZ|(INFO_PALETTE<<12);
  
  // x scroll is "reversed", so subtract from 256 to get value we want
  bg_panel.x_scroll = 256 - panelXOffset*8;
  bg_panel.y_scroll = 0;
  UpdateBackground(&bg_panel);
  // re enable the bg..
  REG_DISPCNT |= BG2_ENABLE;
}

// which character to start at, when more than 1 at this location
int g_maxMapChar;

// called when we move the cursor
void clearInfo(void) {
  // clear "tool tip" and pic of whats here
  clearBuffer();
  blit_buffer();
  clear_text();
  moveHighlighter(g_cursorScreen.x,g_cursorScreen.y);
  
  remove_sprite(0);
  clearSpriteRAM(0);
  clear_objects();
  
  drawPanelBG(0,0);
  g_maxMapChar = 0;
}

const u8 charMapPosition[6][2] = {
  { 0, 12},
  { 5, 12},
  {10, 12},
  { 0, 16},
  { 5, 16},
  {10, 16},
};

/*
Originally, I planned on showing several chars in 
hw scaled sprites, but it looks better with just 1 in
full size portrait with their name.
*/
void drawLittleCharactersHere(int panelXOffset) {
  if (g_total_here != 0 ) {
    
    int gfx, i;
    g_colourType = COLOURTYPE_INFO;
    for (i = 0; i + g_maxMapChar < g_total_here && i < MAX_CHARS_TO_SHOW; i++) {
      gfx = (g_characters[g_charactersHere[i+g_maxMapChar]].race_orders&0xf);
      addUppercaseToken();
      createCharacterTitle(g_charactersHere[i+g_maxMapChar]);
      
      // OBJECT_SPRITE is added to the sprite index, so hack with i - O_S
      drawGFX(gfx, 0,0, i-OBJECT_SPRITE);
      
      MoveSprite(&sprites[i],
                  panelXOffset*8 + 8*charMapPosition[i][0], //x
                  8*(charMapPosition[i][1] - objectGFXTable[gfx]->height)); //y
      
      print_tokens(panelXOffset+4,CHAR_NAME_Y,
                   panelXOffset+4,panelXOffset==0?0xf:0x1e, 
                   FONT_RED_B);
    }
    
    CopyOAM();
  }
}

// create the little red box
void createMapCursor(void) {
  
  u8 tileBuffer[8*8];
  int i, j;
  for (i = 0; i < 8; i++)
    for (j = 0; j < 8; j++) 
      tileBuffer[i+j*8] = (j==0||j==7||i==0||i==7)?MAP_CURSOR_COLOUR:0;
  
  drawImage(tileBuffer,0, 0, MAP_CURSOR_SPRITE);

  // priority above map, below panel
  init_sprite64(MAP_CURSOR_SPRITE, 1);
  // need to say the menu highlighter is on:
  extern int g_highlightOn;
  g_highlightOn = 1;
  
}

// display the panel with the character and terrain feature 
// for the current cursor location
void displayInfoPanel(int leftOrRight) {
  locdat_t loc;
  loc.x = g_cursorLocation.x;
  loc.y = g_cursorLocation.y;
  fillLocationData(&loc);
  
  
  // turn off terrain feature until it is drawn
  REG_DISPCNT &= ~BG0_ENABLE;
  // do we know what's here?
  if (HAVE_EXPLORED_AREA(g_cursorLocation.x+g_cursorLocation.y*64)) {
    drawFeature(loc.terrain, 1, 
                  MAP_FEATURE_X_POS, MAP_FEATURE_Y_POS);
    blit_buffer();
    wait_vsync_int();
    setup_view();
    REG_DISPCNT &= ~BG0_ENABLE;
  }
  
  int panelOffset;
  if (leftOrRight) {
    // on the left
    panelOffset = 0;
    bg_level.x_scroll = 102;
    print_tokens(panelOffset+1,1,panelOffset+1,0xe, FONT_RED_B);
  } else {
    // on the right
    panelOffset = 15;
    print_tokens(panelOffset+1,1,panelOffset+1,0x1d, FONT_RED_B);
    bg_level.x_scroll = (-18)&0xff;
  }
  
  drawPanelBG(panelOffset, 1);
  
  // only set the terrain feature visible if we know what is here:
  if (HAVE_EXPLORED_AREA(g_cursorLocation.x+g_cursorLocation.y*64)) {
    
    bg_level.y_scroll = (-78)&0xff;
    UpdateBackground(&bg_level);
    REG_DISPCNT |= BG0_ENABLE;
  }
  // maybe we are underground...
  drawLittleCharactersHere(panelOffset);
  
}



// set up and display the info panel
void displayInfo(void) {
  // show pic of what is here, and show area name. Only if explored though.
  if (HAVE_EXPLORED_AREA(g_cursorLocation.x+g_cursorLocation.y*64)) {
    addUppercaseToken();
    addTerrainToken(&g_cursorLocation);
  } else {
    // don't know about the surface:
    if (HAVE_EXPLORED_TUNNEL(g_cursorLocation.x+g_cursorLocation.y*64)) {
      addToTokenList(anUndergroundTunnel);
    }
    else {
      addToTokenList(theIcemark);
    }
  }
  
  getAllFriendsAt(&g_cursorLocation);
  displayInfoPanel(g_cursorScreen.x<15?0:1);
//  moveHighlighter(31,21);
  
}

// initialise the cursor and set the start map position
// to the top left corner
void initMapCursor(void) 
{
  createMapCursor();
  g_cursorScreen.x = 0;
  g_cursorScreen.y = 0;
  g_cursorLocation.x = 0;
  g_cursorLocation.y = 0;
  moveHighlighter(g_cursorScreen.x,g_cursorScreen.y);
}

// move the cursor and check if scrolling is needed:
void moveMapCursorUp(void){
  if (g_cursorScreen.y > 0) {
    g_cursorScreen.y--;
    g_cursorLocation.y--;
  } else {
    scroll_up();
    g_cursorLocation.y--;
  }
  if ((s8)g_cursorLocation.y < 0)
    g_cursorLocation.y = 0;
    
  
}

void moveMapCursorDown(void){
  if (g_cursorScreen.y < 19) {
    g_cursorScreen.y++;
    g_cursorLocation.y++;
  } else {
    scroll_down();
    g_cursorLocation.y++;
  }
  if (g_cursorLocation.y >= MAP_HEIGHT )
    g_cursorLocation.y = MAP_HEIGHT-1;
}

void moveMapCursorLeft(void){
  if (g_cursorScreen.x > 0) {
    g_cursorScreen.x--;
    g_cursorLocation.x--;
  } else {
    scroll_left();
    g_cursorLocation.x--;
  }
  if ((s8)g_cursorLocation.x < 0)
    g_cursorLocation.x = 0;
    
}

void moveMapCursorRight(void){
  if (g_cursorScreen.x < 29) {
    g_cursorScreen.x++;
    g_cursorLocation.x++;
  } else {
    scroll_right();
    g_cursorLocation.x++;
  }
  if (g_cursorLocation.x >= MAP_WIDTH )
    g_cursorLocation.x = MAP_WIDTH-1;
  
}

void finishMapScreen(void){
  // clear map and set things as they were
  
  fade_down();
  REG_DISPCNT &= ~BG3_ENABLE;
  clearInfo();
  clearHighlighter();
  init_backgrounds();
  fade_up();
  
}

// use the passed in functions to first scroll one way then return
// amount is the max amount to move (central position)
// limiter is the pointer to x or y, maxVal the maximum size in that dimension
void offsetPosition(void (*moveOneWay)(void), void (*moveBackAgain)(void),
                    int amount,
                    u8 * limiter, u8 maxVal) 
{
  int extra = 0;
  int i;
  for (i = 0; i < amount && (*limiter) < maxVal && (s8)(*limiter) > 0; i++) {
    extra++;
    moveOneWay();
  }
  for (i = 0; i < extra; i++){
    moveBackAgain();
  }
}

void scrollToCharacterPosition(void) {
  // scroll to current char's location...
  int i;
  for (i = 0; i < g_characters[g_saveArea.selectedChar].location.x; i++) {
    moveMapCursorRight();
  }
  for (i = 0; i < g_characters[g_saveArea.selectedChar].location.y; i++) {
    moveMapCursorDown();
  }
  
  // if we are at the bottom of the screen it is confusing; scroll a bit more
  // want to get it so that our x position is 15 from the right edge
  // if x > 15, move right. if x < 15, move left
  if (g_cursorScreen.x > 15) {
    offsetPosition(moveMapCursorRight, moveMapCursorLeft, 15,
                  &g_cursorLocation.x, MAP_WIDTH - 1);
  } else if (g_cursorScreen.x < 15) {
    offsetPosition(moveMapCursorLeft, moveMapCursorRight, 15,
                  &g_cursorLocation.x, MAP_WIDTH - 1);
  }
  
  // and our y position is 10 from the top edge
  // if y > 10, move down. if y < 10, move up
  if (g_cursorScreen.y < 10) {
    offsetPosition(moveMapCursorUp, moveMapCursorDown, 10,
                  &g_cursorLocation.y, MAP_HEIGHT - 1);
    
  } else if (g_cursorScreen.y > 10) {
    offsetPosition(moveMapCursorDown, moveMapCursorUp, 10,
                  &g_cursorLocation.y, MAP_HEIGHT - 1);
  }
  
}

void initMapScreen(void) {
  fade_down();
  clear_text();
  clearHighlighter();
  initMapCursor();
  loadMapTiles();
  
  scrollToCharacterPosition();
  
  clearInfo();
  fade_up();
  // set alpha blend reg to blend between layers (blend mode 1):
  // second: the map (bg3)
  // first:  the panel (bg2)
  //         the text  (bg1)    <--- no effect due to priority?
  //         the terrain (bg0)  <-/
  REG_BLDMOD = BLDMOD_FIRST_BG0 | BLDMOD_FIRST_BG1 | 
               BLDMOD_FIRST_BG2 | BLDMOD_SECOND_BG3 |  
               BLDMOD_MODE(1);
  // blend ratio of 3/4, no brightening 
  REG_COLEV = (13)|((3)<<8);
}

// keep polling key presses until B is pressed
void handleMapKeys(void) {
  int timer = 0;
  int delay = 0;
  timer = TIMER_DELAY;
  // mini key handler:
  while(1) {
    UpdateKeys();
    wait_vsync_int();
    
    delay = (g_cursorScreen.y == 0)?0:MAP_MOVE_DELAY;
    if (KeyPressedNoBounce(KEY_UP,delay)) {
      moveMapCursorUp();      
      if (KeyPressed(KEY_LEFT)) {
        moveMapCursorLeft();
      } else if (KeyPressed(KEY_RIGHT)) {
        moveMapCursorRight();
      }
      
      timer = TIMER_DELAY;
      clearInfo();
      continue;
    }
    
    delay = (g_cursorScreen.y == 19)?0:MAP_MOVE_DELAY;
    if (KeyPressedNoBounce(KEY_DOWN,delay)) {
      if (KeyPressed(KEY_LEFT)) {
        moveMapCursorLeft();
      } else if (KeyPressed(KEY_RIGHT)) {
        moveMapCursorRight();
      }
      moveMapCursorDown();  
      timer = TIMER_DELAY;
      clearInfo();
      continue;
    }
    
    delay = (g_cursorScreen.x == 0)?0:MAP_MOVE_DELAY;
    if (KeyPressedNoBounce(KEY_LEFT,delay)) {
      moveMapCursorLeft();
      if (KeyPressed(KEY_UP)) {
        moveMapCursorUp();
      } else if (KeyPressed(KEY_DOWN)) {
        moveMapCursorDown();
      }
      timer = TIMER_DELAY;
      clearInfo();
      continue;
    }
    
    delay = (g_cursorScreen.x == 29)?0:MAP_MOVE_DELAY;
    if (KeyPressedNoBounce(KEY_RIGHT,delay)) {
      if (KeyPressed(KEY_UP)) {
        moveMapCursorUp();
      } else if (KeyPressed(KEY_DOWN)) {
        moveMapCursorDown();
      }
      moveMapCursorRight();
      timer = TIMER_DELAY;
      clearInfo();
      continue;
    }
    
    delay = 0;
    if (KeyPressedNoBounce(KEY_B,delay)) {
      break;
    }
    
    delay = MAP_A_DELAY;
    if (KeyPressedNoBounce(KEY_A,delay)) {
      timer = TIMER_DELAY;
      if (g_maxMapChar < g_total_here-1 && g_total_here > MAX_CHARS_TO_SHOW) {
        g_maxMapChar++;
        timer = 1;
      } else {
        g_maxMapChar = 0;
      }
      
      
      int tmp = g_maxMapChar;
      clearInfo();
      g_maxMapChar = tmp;
    }
    
    if (timer) {
      timer--;
      if (timer == 0) {
        displayInfo();
      }
    }
  }
  
}

/* 
  Show the map. 
  A timer is set so that if you sit on a square for a bit, 
  it shows a "float over" picture of the terrain, the name 
  and the people at the current cursor position
*/ 
void showMapScreen(void) 
{
  initMapScreen();
  handleMapKeys();
  finishMapScreen();
  
}

