/*
* $Log$
* Revision 2.3  2004/02/22 15:03:54  richq
* Transparency for map screen
*
* Revision 2.2  2004/02/18 19:14:28  richq
* Draws the creatures and characters ahead of us. Apart from some extra
* attribute fiddling, is unchanged from LoM. Also has routines for handling
* the effect of "objects" - like Claws of Death, Wild Horses, etc.
*
* Revision 2.1.1.1  2004/02/04 21:49:21  Default
* Doomdark's Revenge GBA
*                                                                   
*/

#include "gba.h"
#include "keypad.h"

#include "gfx.h"
#include "sprite.h"
#include "objects.h"
#include "objectdata.h"
#include "location.h"
#include "screenmode.h"
#include "tokens.h"
#include "tokendata.h"
#include "look.h"
#include "characters.h"
#include "main.h"
#include "flames.h"
#include "army.h"
#include "fight.h"
#include "savearea.h"

#define OBJECT_PAL_OFFSET 17

// 0xd0 is the last tile - size is therefore 0xd1*8 = 
#define objectGfxData _binary_objectgfx_bin_start
extern u8 objectGfxData[];

// only 6 objects allowed in DDR
const u8 objPosition[] = {
   0x0b,//1
   0x10,//2 
   0x06,//3
   0x15,//4
   0x01,//5
   0x1a //6
};


// things that may appear at a location:
const u8 nastiesTable[16][4] = {
   {0,3,1,4   },  /*Plains    */
   {0,0,2,1   },  /*Mountains */
   {0,3,1,1   },  /*Forest    */
   {0,2,1,1   },  /*Hills     */
   {0,7,8,10  },  /*Gate      */
   {8,9,10,11 },  /*Temple    */
   {2,1,9,12  },  /*Pit       */
   {0,4,7,10  },  /*Palace    */
   {0,1,6,6   },  /*Fortress  */
   {2,6,6,10  },  /*Hall      */
   {3,1,6,6   },  /*Hut       */
   {5,5,5,5   },  /*Tower     */
   {0,6,6,12  },  /*City      */
   {3,12,12,12},  /*Fountain  */
   {0,8,8,11  },  /*Stones    */
   {0,3,3,2   }   /*Fr. Wastes*/
};
  
int g_colourType;

void set_object_palette(void) {
  int i;
  for (i = 0 ; i < 8; i++) {
    OBJPaletteMem[i+OBJECT_PAL_OFFSET] = nonBrightPal[i];
    OBJPaletteMem[i+16+OBJECT_PAL_OFFSET] = brightPal[i];
  }
  
}

// calculate the new colour based on time of day and the given colour
// returns the "transparent" colour, resets *colour to the one to use
// colour format: lower 3 bits == pen. next 3 bits == paper, bit 6 = bright flag
int calculateColourAttrib(int * colour) {
  // 5f04
  if (g_colourType == COLOURTYPE_DAY) {
    // use the paper as transp
    return ((*colour)&0x38)>>3;
    //return colour;
  }
  
  int transp;
  int low = (*colour)&7;
  int high = (*colour)&0x38;
  
  if (g_colourType == COLOURTYPE_DAWN) {
    // dawn...
    if (low == 2 || low == 3)  // red/purple
      low = 0;  // black
    
    if (high == 0x28) // cyan
      high = 0x10;  // red
    transp = 0x10>>3;
    // and use non bright colours...
  } else if (g_colourType == COLOURTYPE_INFO || g_colourType == COLOURTYPE_VIEW) {
    // info screen
    if (low == 7)  // white
      low = 0;  // black
    
    if (high == 0x28) // cyan
      high = 0x38;  // white
    transp = 0x38>>3;
    
    if (g_colourType == COLOURTYPE_VIEW)
      high |= 0x40;
    
    // and use non bright colours...  
  } else if (g_colourType == COLOURTYPE_NIGHT) {
    // night
    // and use bright colours.
    *colour = 0x48;// black/blue
    return 1;  // use blue as te transparent colour
  } else {
    // tunnel
    // use bright colours...
    *colour = 0x42; // red/black
    return 0;  // transparent == black
  }
  
  *colour = (high|low);
  return transp;
}


void drawGFX(int id, int x, int y, int sprite) {
  
  //OBJPaletteMem[7+OBJECT_PAL_OFFSET] = BGPaletteMem[18];
  
  const objectgfx_t * theGFX = objectGFXTable[id];
  int i, j;
  int mapIndex = 0;
  int colourIndex = 0;
  int tile = 0;
  int colour0, colour1;
  u8 tileBuffer[8*8];
  
  gfxTable = objectGfxData;
  int paper, ink;
  int colourAttrib;
  int useTransparent;
  int brightOffset = 0;
  for (j = 0; j < theGFX->height; j++) {
    for (i = 0; i < theGFX->width; i++) {
      tile = theGFX->tileMap[mapIndex];
      if (tile != 0) {
        colourAttrib = theGFX->colourMap[colourIndex];
        useTransparent = calculateColourAttrib(&colourAttrib);
        colour0 = colourAttrib&0x7;
        colour1 = (colourAttrib>>3)&0x7;
        if (colourAttrib & 0x40)
          brightOffset = 16;
        else
          brightOffset = 0;
        paper = (colour1==useTransparent && 
              (j==0 || i==0 || i==theGFX->width-1 || 
                                  g_status.screen == MENU_SCREEN) )?0
               : colour1+OBJECT_PAL_OFFSET+brightOffset;
        ink = colour0+OBJECT_PAL_OFFSET+brightOffset;
        
        
        
        createTile( tileBuffer, tile, ink, paper);
        drawImage(tileBuffer, i, j, OBJECT_SPRITE+sprite);
        
      }
      mapIndex++;
      colourIndex++;
    }
  }
  
  init_sprite64(OBJECT_SPRITE+sprite, 0);
  MoveSprite(&sprites[OBJECT_SPRITE+sprite], x*8, y*8);
}

int previousType;
int maxAllowedObjects;
void drawObject(int i, int y, int *amount){
  if (*amount < maxAllowedObjects) {
    int x = objPosition[*amount];
    if (previousType == i) {
      int sprite = OBJECT_SPRITE + (*amount);
      init_sprite64(sprite, 0);
      sprites[sprite].attribute2 = sprites[sprite-1].attribute2;
      MoveSprite(&sprites[sprite], x*8, y*8);
    }
    else {
      previousType = i;
      drawGFX(i, x, y, *amount);
    }
    (*amount)++;
  }
}


// 655d
void drawCharactersHere(int * amount) {

  int i;
  int gfx;
  for (i = 0; i < g_total_here; i++) {
    gfx = (g_characters[g_charactersHere[i]].race_orders&0xf);
    drawObject( gfx,
                0x14-objectGFXTable[gfx]->height, 
                amount);
  }
  
}


/*
drawInFront
 draw whatever is infront of us. 
 
 based on 6510.
 
 colourtype used to determine palette offsets
 for the objects (i.e dark at night, silhouetted in the morning, red in tunnels)

; 0 for day.
; 1 for dawn
; 2 for info screen
; 3 for night
; 4 for main view
; 5 for tunnel
 
*/
void drawInFront(int colourType) {
  
  if (!g_status.canDrawInfront)
    return;
  
  g_colourType = colourType;
  
  set_object_palette(/*g_colourType*/);
  
  int amount = 0;
  previousType = -1;
  maxAllowedObjects = MAX_OBJECTS - 1;
  if (g_status.charInTunnel) {
    maxAllowedObjects = 3;
  }
  
  
  if (g_total_here != 0) {
    // folk to draw.
    drawCharactersHere(&amount);
  }
  
  if (g_localDetails.thing_infront <= OBJECT_WILD_HORSES) {
    // things to draw.
    int j;
    int gfx = g_localDetails.thing_infront + 16;
    for (j = 0; j < 4; j++) 
      drawObject(gfx,
                 0x14-objectGFXTable[gfx]->height,
                 &amount);
  }
  
  // display them
  CopyOAM();
  
}


/*
Draws the character graphics at the given (absolute) x, y
Uses the old "SHIELD_SPRITE" trick, which is probably sprite 0
*/
void drawCharGfx(int glyph, int x, int basey, int colourType) {
  g_colourType = colourType;
  
  int y = (basey-(objectGFXTable[glyph]->height*8))&255;
  
  set_object_palette(/*g_colourType*/);
  
  drawGFX(glyph, 16,20,SHIELD_SPRITE);
  
  init_sprite64(OBJECT_SPRITE+SHIELD_SPRITE, 2);
  move_sprite(SHIELD_SPRITE+OBJECT_SPRITE, x, y);
  
}


void clear_objects(void) {
  int i;
  for (i = 0; i < MAX_OBJECTS; i++) {
    remove_sprite(OBJECT_SPRITE + i);
  }
  CopyOAM();
  
  // clear vram too
  for (i = 0; i < MAX_OBJECTS-2; i++) {
    clearSpriteRAM(OBJECT_SPRITE + i);
  }
  
}


/*
Calculate the objects at a location
based on code at 7105
returns object id
*/
int calculateObject(const location_t * pos) {

    int thing_here = OBJECT_EMPTY;
    
    if ( !HAS_OBJECT(g_locdat.info) ) {
      return thing_here;
    }
    int terrainToUse = g_locdat.terrain;
    if ( HAS_TUNNEL(g_locdat.info) && g_locdat.terrain < GATE) {
      // is a "passageway" tunnel, not an entrance
      // passageways can't have things on the surface...
      if (!g_status.charInTunnel)
        return thing_here;
      terrainToUse = WASTES;
    } else {
      // is a non-tunnel or a tunnel entrance. 
      // Either way, being inside a tunnel == can't see this square's object
      if (g_status.charInTunnel)
        return thing_here;
      // if the surface is frozen wastes, then there is nothing there
      if (g_locdat.terrain == WASTES)
        return thing_here;
    }
    
    // terrain naming does this
    int terrainNameVal = calculateNameSeed(pos->x, pos->y)&3;
    
    thing_here = nastiesTable[terrainToUse][terrainNameVal]; // 7147 LD   HL,7151   ; nasties table
    return thing_here;
    
}


//836f
void getHorse(void) {  
//  if ( (g_selectedChar->race_orders & 0xf) < 12) {
//    g_selectedChar->race_orders &= ~0x1; // clears lower bit
//  }  
  if ( (g_characters[g_saveArea.selectedChar].race_orders & 0xf) < RACE_TARITHEL) {
    g_characters[g_saveArea.selectedChar].race_orders &= ~0x1; // clears lower bit
  }  
  
}

void sheltered(void) {
  // update tiredness
//  int newval = g_selectedChar->energy + 0x20;
//  g_selectedChar->energy = (newval>0xff)?0xff:newval;
  int newval = g_characters[g_saveArea.selectedChar].energy + 0x20;
  g_characters[g_saveArea.selectedChar].energy = (newval>0xff)?0xff:newval;
}



void activateObject(int thing_here) {
  switch (thing_here) {
    case OBJECT_WILD_HORSES:  getHorse(); 
                              return;
    case OBJECT_GUIDANCE   :  return;
    case OBJECT_SHELTER    ://837b
                            sheltered(); 
                            return;
    case OBJECT_CLAWS      :// make night time://8385
                            g_characters[g_saveArea.selectedChar].time_looking 
                            = g_characters[g_saveArea.selectedChar].time_looking & 0x7;
                            return;
    case OBJECT_FLAMES     :// make dawn://8388
                            g_characters[g_saveArea.selectedChar].time_looking 
                            = (DAWN<<3) | 
                              (g_characters[g_saveArea.selectedChar].time_looking & 0x7);
                            return;
    case OBJECT_THORNS     ://838d
                            g_characters[g_saveArea.selectedChar].despondence = 0;
                            return;
    case OBJECT_BLOOD      ://8390
                            g_characters[g_saveArea.selectedChar].despondence = 0xff;
                            return;
    case OBJECT_LANGUOR    ://8395
                            g_characters[g_saveArea.selectedChar].energy = 0;
                            return;
    case OBJECT_SPRINGS    ://8398
                            g_characters[g_saveArea.selectedChar].energy = 0xff;
                            return;
    default: break;
  }
  
  // must be a baddie...
  // lose some troops, randomly get killed. etc.
  int r = randomish()&15;
  decreaseU8By(&g_charArmies[g_saveArea.selectedChar].size, r/2);

///*debug mode - press down to always kill person*/
//UpdateKeys();
//if (!KeyPressed(KEY_DOWN) ){  
  if ( (thing_here + 8) >= r) {
    return;
  }
//}

  fightResults(thing_here+128, g_saveArea.selectedChar, r);
  
}

void remove_shield(void) {
  remove_sprite(OBJECT_SPRITE+SHIELD_SPRITE);
  clearSpriteRAM(OBJECT_SPRITE+SHIELD_SPRITE);
//  g_lastShield = -1;
}


void describeObjects(void) {
  
  if (g_localDetails.thing_here >= OBJECT_NOTHING) {
    addSingleToken(0x9b);
    return;
  } 
  
  if (g_localDetails.thing_here >= OBJECT_CLAWS)
    addToTokenList(the_);
  addSingleToken(0x7b + g_localDetails.thing_here);
  
  if (g_localDetails.thing_here < OBJECT_WILD_HORSES) {
    addToTokenList(andSlainThem);
  }
  else {
    if (g_localDetails.thing_here == OBJECT_GUIDANCE) {
      describeGuidance();
    }
    else if (g_localDetails.thing_here >= OBJECT_CLAWS) {
      addToTokenList(of_);
      addSingleToken(0x81 + g_localDetails.thing_here);
    }
    
    
  }
  
  
}

