/*
* $Log$
* Revision 2.2  2004/02/18 19:13:23  richq
* Creates the list of controllable characters. Up to 127 of them! Similar to the
* LoM code, but with the addition of "pages" of characters. L+R cycle pages, up
* and down move the cursor. Oh, and dead characters are shown in black.
*
* Revision 2.1.1.1  2004/02/04 21:49:21  Default
* Doomdark's Revenge GBA
*
*
*/

#include "gfx.h"
#include "select.h"
#include "look.h"
#include "characters.h"
#include "infoscreen.h"
#include "savearea.h"
#include "tokens.h"
#include "tokendata.h"
#include "menu.h"
#include "main.h"


#define MAX_LOYALS_SHOWN ((160/8) - 6)

int g_hiliteSelect;

//const char * keyboardContols = "cvbn1234567890qwertyuiopasdfghjk";
int g_selectableThings;
u8 g_selectableThingTable[128];

typedef struct {
  u8 prev;
  u8 current;
  u8 next;
}
listindex_t;

listindex_t g_indices;

const char selectKeys[] = {
/*0x73,*/
0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,    /*1-8*/
0x39,0x30,0x71,0x77,0x65,0x72,0x74,0x79,    /*90qwerty*/
0x75,0x69,0x6f,0x70                         /*uiop*/
};



void displaySelectableCharList(int offset){
  char control;
  int i;
  int colour;
  int charid;
  
  if (offset < g_selectableThings) {
    clear_text();
  } else {
    return;
  }
  g_indices.current = offset;
  int previous = 1 + offset - MAX_LOYALS_SHOWN;
  g_indices.prev = previous<0?0:previous;
  
  g_hiliteSelect = 0;
  for (i = offset; i  < g_selectableThings && (i-offset) < MAX_LOYALS_SHOWN; i++) {
    control = ASCII_BOX_OK;
    charid = g_selectableThingTable[i];
    if (charid == g_saveArea.selectedChar)
      g_hiliteSelect = i-offset;
    
    if (TIME_LEFT(g_characters[charid].time_looking) == 0) {
      control = ASCII_BOX_NIGHT;
    } else if (g_characters[charid].energy < 0x20) {
      control = ASCII_BOX_TIRED;
    }
    
    if (g_characters[charid].reckless == 0) {
      colour = FONT_BLACK;
      control = ASCII_BOX_DEAD;
    } else {
      colour = FONT_RED;
    }
    
    addNewlineToken();
    addLiteralToken(control);
    addLiteralToken(selectKeys[i-offset]);
    createCharacterTitle(charid);
    
    if (i-offset == 0) {
      print_tokens(0, 1, 0, 0x1e, colour);
    } else {
      print_tokens_pal(colour);
    }
  }
  
  if (i < g_selectableThings || offset != 0) {
    addMoreAndPrint();
    g_indices.next = i-1;
  }
}


void createSelectableCharList(void) {
  int i;
  g_selectableThings = 0;
  g_hiliteSelect = 0;
  
  for (i = 0; i< 128; i++) {
    if (LOYALTY(g_characters[i].flags) == 0) {
      if (i == g_saveArea.selectedChar)
        g_hiliteSelect = g_selectableThings;
      g_selectableThingTable[g_selectableThings++] = i;
    }
  }
  
}


// 739b
void select(void) {
  
  setupInfoScreen();
  createSelectableCharList();
  
  displaySelectableCharList(0);
  
  // now have a list both on screen and in mem of the selectable chars
  // originally, the player was forced to choose.. but here it
  // works like any other screen.
  selectOption(g_hiliteSelect);

  g_status.screen = SELECT_SCREEN;
  
}


void showNextSelectPage(void) {
  displaySelectableCharList(g_indices.next);
  selectOption(g_hiliteSelect);
}

void showPrevSelectPage(void) {
  displaySelectableCharList(g_indices.prev);
  g_hiliteSelect = g_indices.next - g_indices.current;
  selectOption(g_hiliteSelect);
}

void selectNextChar(void) {
  
  int character = g_saveArea.selectedChar;
  while (character < 127) {
    character++;
    
    if (LOYALTY(g_characters[character].flags) == 0) {
      g_saveArea.selectedChar = character;
      //setSelectedChar(g_saveArea.selectedChar);
      //set_palette_time(TIME_LEFT(g_selectedChar->time_looking));
      
      g_status.screen = LOOK_SCREEN+1;
      //set_palette_time(TIME_LEFT(g_characters[g_saveArea.selectedChar].time_looking));
      look();
      break;
    }
  }
  
}

void selectPrevChar(void) {
  
  int character = g_saveArea.selectedChar;
  while (character > 0) {
    character--;
    
    if (LOYALTY(g_characters[character].flags) == 0) {
      g_saveArea.selectedChar = character;
      //setSelectedChar(g_saveArea.selectedChar);
      //set_palette_time(TIME_LEFT(g_selectedChar->time_looking));
      g_status.screen = LOOK_SCREEN+1;
      look();
      break;
    }
  }
  
}


void makeSelection(void){
  clearHighlighter();
  g_saveArea.selectedChar = g_selectableThingTable[g_hiliteSelect + g_indices.current];
  //setSelectedChar(g_saveArea.selectedChar);
  //set_palette_time(TIME_LEFT(g_selectedChar->time_looking));
  //set_palette_time(TIME_LEFT(g_characters[g_saveArea.selectedChar].time_looking));
  look();
  
}

void selectDown(void){
  g_hiliteSelect++;
  
  if (g_hiliteSelect < MAX_LOYALS_SHOWN && 
      (g_hiliteSelect + g_indices.current) < g_selectableThings ) 
  {  
    selectOption(g_hiliteSelect);
  } else if ((g_hiliteSelect + g_indices.current) < g_selectableThings) {
    // run to end of list
    showNextSelectPage();
  }
  else {
    g_hiliteSelect--;
  }
  
}

void selectUp(void){
  if (g_hiliteSelect) {
    g_hiliteSelect--;
    selectOption(g_hiliteSelect);
  } else if (g_indices.current != 0) {
    showPrevSelectPage();
  }
  
}


